// license:BSD-3-Clause
// copyright-holders:Hiromasa Tanaka
import { WASI } from '@wasmer/wasi';
import { WasmFs } from "@wasmer/wasmfs";
import { lowerI64Imports } from "@wasmer/wasm-transformer";
import browserBindings from '@wasmer/wasi/lib/bindings/browser';
import { spy } from 'spyfs';

// wasi instance
export let wasi;
export let wasmFs;
export let memFs;

/**
 * Initialize WebAssembly with wasmer-js
 *
 * Need wasm-bindgen generate source code patch for insert wasm export.
 * A patch is needed to insert wasmer-js WASI instance.
 *
 * > import * as wasm from './libymfm_bg.wasm';
 * < let wasm; export function setWasmExport(exports) { wasm = exports; }
 *
 * @see scripts/wasm_bindgen_patch.js
 * @returns instance.exports
 */
export async function initWasi() {
    // memfs + spy
    wasmFs = new WasmFs();
    // memfs not working ?
    memFs = spy(wasmFs.fs, async (action) => {
        console.log({ [action.method] : {
            "isAsync": action.isAsync,
            "args": action.args,
        }});
        await action;
    });
    // create WASI instance
    wasi = new WASI({
        args: [""],
        env: {},
        bindings: {
            ...browserBindings,
            fs: memFs
        }
    });
    // fetch wasm module
    const response = await fetch(new URL('../wasm/libymfm_bg.wasm', import.meta.url));
    const responseArrayBuffer = new Uint8Array(await response.arrayBuffer());
    // compile wasm
    const wasm_bytes = new Uint8Array(responseArrayBuffer).buffer;
    const lowered_wasm = await lowerI64Imports(wasm_bytes);
    let module = await WebAssembly.compile(lowered_wasm);
    // get WASI imports
    let imposts = wasi.getImports(module);
    // merge wasm imports
    //   (import "wasi_snapshot_preview1" "fd_seek" (func $__wasi_fd_seek (type $t25)))
    //   (import "./libymfm_bg.js" "__wbg_new_59cb74e423758ede"...)
    imposts['./libymfm_bg.js'] = await import('../wasm/libymfm_bg');
    const instance = await WebAssembly.instantiate(module, {
        ...imposts
    });
    // start wasi
    wasi.start(instance);

    // return wasm exports(for call setWasmExport())
    return instance.exports;
}
